@@ -11,7 +11,7 @@ import numpy as np
 
 try:
     from scipy.ndimage import zoom
-    from scipy.interpolate import interp1d
+    from scipy import stats
     HAVE_SCIPY = True
 except ImportError:
     HAVE_SCIPY = False
@@ -47,7 +47,7 @@ COLUMNAR_DATA_SOURCES = tuple(COLUMNAR_DATA_SOURCES)
 
 from operator import itemgetter
 
-__version__ = '0.5.2'
+__version__ = '0.6.3'
 
 
 class CoordinateOutOfRangeException(Exception):
@@ -75,48 +75,16 @@ class MultiHistBase(object):
     def __setitem__(self, key, value):
         self.histogram[key] = value
 
+    # Let unary operators work on wrapped histogram:
+    def min(self):
+        return self.histogram.min()
+
+    def max(self):
+        return self.histogram.max()
+
     def __len__(self):
         return len(self.histogram)
 
-    def __add__(self, other):
-        return self.__class__.from_histogram(self.histogram.__add__(other), self.bin_edges, self.axis_names)
-
-    def __sub__(self, other):
-        return self.__class__.from_histogram(self.histogram.__sub__(other), self.bin_edges, self.axis_names)
-
-    def __mul__(self, other):
-        return self.__class__.from_histogram(self.histogram.__mul__(other), self.bin_edges, self.axis_names)
-
-    def __truediv__(self, other):
-        return self.__class__.from_histogram(self.histogram.__truediv__(other), self.bin_edges, self.axis_names)
-
-    def __floordiv__(self, other):
-        return self.__class__.from_histogram(self.histogram.__floordiv__(other), self.bin_edges, self.axis_names)
-
-    def __mod__(self, other):
-        return self.__class__.from_histogram(self.histogram.__mod__(other), self.bin_edges, self.axis_names)
-
-    def __divmod__(self, other):
-        return self.__class__.from_histogram(self.histogram.__divmod__(other), self.bin_edges, self.axis_names)
-
-    def __pow__(self, other):
-        return self.__class__.from_histogram(self.histogram.__pow__(other), self.bin_edges, self.axis_names)
-
-    def __lshift__(self, other):
-        return self.__class__.from_histogram(self.histogram.__lshift__(other), self.bin_edges, self.axis_names)
-
-    def __rshift__(self, other):
-        return self.__class__.from_histogram(self.histogram.__rshift__(other), self.bin_edges, self.axis_names)
-
-    def __and__(self, other):
-        return self.__class__.from_histogram(self.histogram.__and__(other), self.bin_edges, self.axis_names)
-
-    def __xor__(self, other):
-        return self.__class__.from_histogram(self.histogram.__xor__(other), self.bin_edges, self.axis_names)
-
-    def __or__(self, other):
-        return self.__class__.from_histogram(self.histogram.__or__(other), self.bin_edges, self.axis_names)
-
     def __neg__(self):
         return self.__class__.from_histogram(-self.histogram, self.bin_edges, self.axis_names)
 
@@ -129,8 +97,30 @@ class MultiHistBase(object):
     def __invert__(self):
         return self.__class__.from_histogram(~self.histogram, self.bin_edges, self.axis_names)
 
+    # Let binary operators work on wrapped histogram
+
+    @classmethod
+    def _make_binop(cls, opname):
+        def binop(self, other):
+            return self.__class__.from_histogram(
+                getattr(self.histogram, opname)(other),
+                self.bin_edges,
+                self.axis_names)
+        return binop
+
+for methodname in 'add sub mul div truediv floordiv mod divmod pow lshift rshift and or'.split():
+    dundername = '__%s__' % methodname
+    setattr(MultiHistBase,
+            dundername,
+            MultiHistBase._make_binop(dundername))
+    setattr(MultiHistBase,
+            '__r%s__' % methodname,
+            getattr(MultiHistBase, dundername))
+
+# Verbose alias
 MultiHistBase.similar_blank_histogram = MultiHistBase.similar_blank_hist
 
+
 class Hist1d(MultiHistBase):
     axis_names = None
     dimensions = 1
@@ -193,11 +183,13 @@ class Hist1d(MultiHistBase):
         cs = np.cumsum(self.histogram)
         return cs / cs[-1]
 
-    def get_random(self, *args, **kwargs):
+    def get_random(self, size=10):
         """Returns random variates from the histogram.
-        TODO: Only bin centers can be returned!
+        Note this assumes the histogram is an 'events per bin', not a pdf.
+        Inside the bins, a uniform distribution is assumed.
         """
-        return np.random.choice(self.bin_centers, p=self.normalized_histogram, *args, **kwargs)
+        bin_i = np.random.choice(np.arange(len(self.bin_centers)), size=size, p=self.normalized_histogram)
+        return self.bin_centers[bin_i] + np.random.uniform(-0.5, 0.5, size=size) * self.bin_volumes()[bin_i]
 
     def items(self):
         """Iterate over (bin_center, hist_value) from left to right"""
@@ -218,29 +210,88 @@ class Hist1d(MultiHistBase):
             bc = 1
         return np.sqrt(np.average((self.bin_centers - self.mean) ** 2, weights=self.histogram)) * bc
 
-    def plot(self, normed=False, scale_errors_by=1.0, scale_histogram_by=1.0, plt=plt, errors=False, **kwargs):
-        """Plots the histogram with Poisson (sqrt(n)) error bars
-          - scale_errors_by multiplies the error bars by its argument
-          - scale_histogram_by multiplies the histogram AND the error bars by its argument
-          - plt thing to call .errorbar on (pylab, figure, axes, whatever the matplotlib guys come up with next)
+    def plot(self,
+             normed=False, scale_histogram_by=1.0, scale_errors_by=1.0,
+             errors=False, error_style='bar', error_alpha=0.3,
+             plt=plt, set_xlim=False,
+             **kwargs):
+        """Plot the histogram, with error bars if desired.
+
+        :param normed: Scale the histogram so the sum is 1 before plotting.
+        Errors are computed before scaling, then scaled accordingly.
+        :param scale_histogram_by: Custom multiplier to apply to histogram.
+        Errors are computed before scaling, then scaled accordingly.
+        :param scale_errors_by: Custom multiplier to apply to errors.
+        :param errors: Whether and how to plot 1 sigma error bars
+            * False for no errors
+            * True or 'fc' for Feldman-Cousin errors.
+              For > 20 events, central Poisson intervals are used
+            * 'central' for central Poisson confidence intervals
+            * 'sqrtn' for sqrt(n) errors
+        :param errorstyle: How to plot errors (if errors is not False)
+         * 'bar' for error bars
+         * 'band' for shaded bands
+        :param error_alpha: Alpha multiplier for errorstyle='band'
+        :param plt: Object to call plt... on; matplotlib.pyplot by default.
+        :param set_xlim: If True, set xlim to the range of the hist
         """
         if not CAN_PLOT:
-            raise ValueError("matplotlib did not import, so can't plot your histogram...")
-        if errors:
-            kwargs.setdefault('linestyle', 'none')
-            yerr = np.sqrt(self.histogram)
-            if normed:
-                y = self.normed_histogram
-                yerr /= self.n
-            else:
-                y = self.histogram.astype(np.float)
-            yerr *= scale_errors_by * scale_histogram_by
-            y *= scale_histogram_by
-            plt.errorbar(self.bin_centers, y, yerr,
-                         marker='.', **kwargs)
+            raise ValueError(
+                "matplotlib did not import, so can't plot your histogram...")
+        x = self.bin_edges
+        y = self.histogram
+        if normed:
+            scale_histogram_by /= y.sum()
+        if errors == 'sqrtn':
+            _yerr = y**0.5
+            ylow, yhigh = y - _yerr, y + _yerr
+        elif errors == 'central':
+            ylow, yhigh = poisson_1s_interval(y, fc=False)
+        elif errors:
+            ylow, yhigh = poisson_1s_interval(y, fc=True)
         else:
-            kwargs.setdefault('linestyle', 'steps-mid')
-            plt.plot(self.bin_centers, self.histogram, **kwargs)
+            ylow, yhigh = y, y
+
+        y = y.astype(np.float) * scale_histogram_by
+        ylow = ylow.astype(np.float) * scale_histogram_by * scale_errors_by
+        yhigh = yhigh.astype(np.float) * scale_histogram_by * scale_errors_by
+
+        if errors and error_style == 'bar':
+            kwargs.setdefault('linestyle', 'none')
+            kwargs.setdefault('marker', '.')
+            plt.errorbar(self.bin_centers,
+                         y,
+                         yerr=[y - ylow, yhigh - y],
+                         **kwargs)
+        else:
+            # Note we use steps-pre, not steps-mid.
+            # If we would have plotted values only vs the centers
+            #  * the steps won't be correct for log scales
+            #  * the final bins will not fully show
+            def fix(q):
+                return np.concatenate([[q[0]], q])
+            y = fix(y)
+            ylow = fix(ylow)
+            yhigh = fix(yhigh)
+
+            if errors and error_style == 'band':
+                plt.plot(x, y, drawstyle='steps-pre', **kwargs)
+                alpha = error_alpha
+                if 'alpha' in kwargs:
+                    alpha *= kwargs['alpha']
+                    del kwargs['alpha']
+                kwargs['linewidth'] = 0
+                if 'label' in kwargs:
+                    # Don't want to double-label!
+                    del kwargs['label']
+                plt.fill_between(x, ylow, yhigh,
+                                 alpha=alpha,
+                                 step='pre', **kwargs)
+            else:
+                plt.plot(x, y, drawstyle='steps-pre', **kwargs)
+
+        if set_xlim:
+            plt.xlim(x[0], x[-1])
 
     def percentile(self, percentile):
         """Return bin center nearest to percentile"""
@@ -272,6 +323,7 @@ class Histdd(MultiHistBase):
         """Make a HistdD from numpy histogram + bin edges
         :param histogram: Initial histogram
         :param bin_edges: x bin edges of histogram, y bin edges, ...
+        :param axis_names: Names of axes
         :return: Histnd instance
         """
         bin_edges = np.array(bin_edges)
@@ -286,7 +338,7 @@ class Histdd(MultiHistBase):
         # dimensions is a shorthand [(axis_name_1, bins_1), (axis_name_2, bins_2), ...]
         if 'dimensions' in kwargs:
             kwargs['axis_names'], kwargs['bins'] = zip(*kwargs['dimensions'])
-    
+
         if len(data) == 0:
             if kwargs['range'] is None:
                 if kwargs['bins'] is None:
@@ -306,12 +358,20 @@ class Histdd(MultiHistBase):
     def add(self, *data, **kwargs):
         self.histogram += self._data_to_hist(data, **kwargs)[0]
 
+    @staticmethod
+    def _is_columnar(x):
+        if isinstance(x, COLUMNAR_DATA_SOURCES):
+            return True
+        if isinstance(x, np.ndarray) and x.dtype.fields:
+            return True
+        return False
+
     def _data_to_hist(self, data, **kwargs):
         """Return bin_edges, histogram array"""
         if hasattr(self, 'bin_edges'):
             kwargs.setdefault('bins', self.bin_edges)
 
-        if len(data) == 1 and isinstance(data[0], COLUMNAR_DATA_SOURCES):
+        if len(data) == 1 and self._is_columnar(data[0]):
             data = data[0]
 
             if self.axis_names is None:
@@ -328,7 +388,9 @@ class Histdd(MultiHistBase):
                 for partition in data.to_delayed():
                     ph = dask.delayed(Histdd)(partition, axis_names=self.axis_names, bins=kwargs['bins'])
                     ph = dask.delayed(lambda x: x.histogram)(ph)
-                    ph = dask.array.from_delayed(ph, shape=fake_histogram.histogram.shape, dtype=fake_histogram.dtype)
+                    ph = dask.array.from_delayed(ph,
+                                                 shape=fake_histogram.histogram.shape,
+                                                 dtype=fake_histogram.histogram.dtype)
                     partial_hists.append(ph)
                 partial_hists = dask.array.stack(partial_hists, axis=0)
 
@@ -341,7 +403,8 @@ class Histdd(MultiHistBase):
                 return histogram, bin_edges
 
             else:
-                data = np.vstack([data[x].values for x in self.axis_names])
+                data = np.vstack([data[x].values if isinstance(data, pd.DataFrame) else data[x]
+                                  for x in self.axis_names])
 
         data = np.array(data).T
         return np.histogramdd(data,
@@ -436,8 +499,9 @@ class Histdd(MultiHistBase):
             # Make a 1=bin slice
             stop = start
         axis = self.get_axis_number(axis)
-        start_bin = self.get_axis_bin_index(start, axis)
-        stop_bin = self.get_axis_bin_index(stop, axis)
+        start_bin = max(0, self.get_axis_bin_index(start, axis))
+        stop_bin = min(len(self.bin_centers(axis)) - 1,  # TODO: test off by one!
+                       self.get_axis_bin_index(stop, axis))
         new_bin_edges = self.bin_edges.copy()
         new_bin_edges[axis] = new_bin_edges[axis][start_bin:stop_bin + 2]   # TODO: Test off by one here!
         return Histdd.from_histogram(np.take(self.histogram, np.arange(start_bin, stop_bin + 1), axis=axis),
@@ -463,10 +527,10 @@ class Histdd(MultiHistBase):
         return Histdd.from_histogram(np.cumsum(self.histogram, axis=axis),
                                      bin_edges=self.bin_edges,
                                      axis_names=self.axis_names)
-                                     
+
     def _simsalabim_slice(self, axis):
         return [slice(None) if i != axis else np.newaxis
-                            for i in range(self.dimensions)]
+                for i in range(self.dimensions)]
 
     def normalize(self, axis):
         """Returns new histogram where all values along axis (in one bin of the other axes) sum to 1"""
@@ -474,7 +538,7 @@ class Histdd(MultiHistBase):
         sum_along_axis = np.sum(self.histogram, axis=axis)
         # Don't do anything for subspaces without any entries -- this avoids nans everywhere
         sum_along_axis[sum_along_axis == 0] = 1
-        hist = self.histogram / sum_along_axis[self._simsalabim_slice(axis)]
+        hist = self.histogram / sum_along_axis[tuple(self._simsalabim_slice(axis))]
         return Histdd.from_histogram(hist,
                                      bin_edges=self.bin_edges,
                                      axis_names=self.axis_names)
@@ -501,22 +565,40 @@ class Histdd(MultiHistBase):
         """
         axis = self.get_axis_number(axis)
 
+        # Shape of histogram
+        s = self.histogram.shape
+
+        # Shape of histogram after axis has been collapsed to 1
+        s_collapsed = list(s)
+        s_collapsed[axis] = 1
+
+        # Shape of histogram with axis removed entirely
+        s_removed = np.concatenate([s[:axis], s[axis + 1:]]).astype(np.int)
+
         # Using np.where here is too tricky, as it may not return a value for each "bin-columns"
-        # First get an array which has a minimum at the percentile-containing bin on each axis
+        # First, get an array which has a minimum at the percentile-containing bins
+        # The minimum may not be unique: if later bins are empty, they will not be
         ecdf = self.cumulative_density(axis).histogram
         if not inclusive:
-            density = self.normalize(axis).histogram
-            ecdf = ecdf - density
-        hist_with_extrema = ecdf - 2 * (ecdf >= percentile / 100)
+            raise NotImplementedError("Non-inclusive percentiles not yet implemented")
+        ecdf = np.nan_to_num(ecdf)    # Since we're relying on self-equality later
+        x = ecdf - 2 * (ecdf >= percentile / 100)
 
-        # Now find the extremum indices using np.argmin
-        percentile_indices = np.argmin(hist_with_extrema, axis=axis)
+        # We now want to get the location of the minimum
+        # To ensure it is unique, add a very very very small monotonously increasing bit to x
+        # Nobody will want 1e-9th percentiles, right? TODO
+        sz = np.ones(len(s), dtype=np.int)
+        sz[axis] = -1
+        x += np.linspace(0, 1e-9, s[axis]).reshape(sz)
 
-        # Finally, convert from extremum indices to bin centers
-        # See first example under 'Advanced indexing' in http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html
-        index = [np.arange(q) for q in self.histogram.shape]
-        index[axis] = percentile_indices
-        result = self.all_axis_bin_centers(axis=axis)[index]
+        # 1. Find the minimum along the axis
+        # 2. Reshape to s_collapsed and perform == to get a mask
+        # 3. Apply the mask to the bin centers along axis
+        # 4. Unflatten with reshape
+        result = self.all_axis_bin_centers(axis)[
+            x == np.min(x, axis=axis).reshape(s_collapsed)
+        ]
+        result = result.reshape(s_removed)
 
         if self.dimensions == 2:
             new_hist = Hist1d
@@ -551,7 +633,7 @@ class Histdd(MultiHistBase):
             average = average[self._simsalabim_slice(axis)]
             variance = np.average((values-average)**2, weights=weights, axis=axis)
             return np.sqrt(variance)
-        
+
         axis = self.get_axis_number(axis)
         std_hist = weighted_std(self.all_axis_bin_centers(axis),
                                 weights=self.histogram, axis=axis)
@@ -567,19 +649,23 @@ class Histdd(MultiHistBase):
     # Other stuff
     ##
     def bin_volumes(self):
-         return reduce(np.multiply, np.ix_(*[np.diff(bs) for bs in self.bin_edges]))
+        return reduce(np.multiply, np.ix_(*[np.diff(bs) for bs in self.bin_edges]))
 
-
-    def rebin(self, *factors, order=1):
+    def rebin(self, *factors, **kwargs):
         """Return a new histogram that is 'rebinned' (zoomed) by factors (tuple of floats) along each dimensions
           factors: tuple with zoom factors along each axis. e.g. 2 = double number of bins, 0.5 = halve them.
           order: Order for spline interpolation in scipy.ndimage.zoom. Defaults to  linear interpolation (order=1).
 
+        The only accepted keyword argument is 'order'!!! (python 2 is not nice)
+
         The normalization is set to the normalization of the current histogram
         The factors don't have to be integers or fractions: scipy.ndimage.zoom deals with the rebinning arcana.
         """
         if not HAVE_SCIPY:
             raise NotImplementedError("Rebinning requires scipy.ndimage")
+        if any([x != 'order' for x in kwargs.keys()]):
+            raise ValueError("Only 'order' keyword argument is accepted. Yeah, this is confusing.. blame python 2.")
+        order = kwargs.get('order', 1)
 
         # Construct a new histogram
         mh = self.similar_blank_histogram()
@@ -597,13 +683,14 @@ class Histdd(MultiHistBase):
         for i, f in enumerate(factors):
             x = self.bin_edges[i]
             mh.bin_edges[i] = np.interp(
-                x=np.linspace(0, 1, (len(x) - 1) * f + 1),
+                x=np.linspace(0, 1, round((len(x) - 1) * f) + 1),
                 xp=np.linspace(0, 1, len(x)),
                 fp=x)
 
         # Rebin the histogram using ndimage.zoom, then renormalize
         mh.histogram = zoom(self.histogram, factors, order=order)
-        mh.histogram *= self.histogram.sum() / mh.histogram.sum()
+        if mh.histogram.sum() != 0:
+            mh.histogram *= self.histogram.sum() / mh.histogram.sum()
         # mh.histogram /= np.product(factors)
 
         return mh
@@ -611,13 +698,15 @@ class Histdd(MultiHistBase):
     def get_random(self, size=10):
         """Returns (size, n_dim) array of random variates from the histogram.
         Inside the bins, a uniform distribution is assumed
-        TODO: Test!
+        Note this assumes the histogram is an 'events per bin', not a pdf.
+        TODO: test more.
         """
         # Sample random bin centers
         bin_centers_ravel = np.array(np.meshgrid(*self.bin_centers(),
                                                  indexing='ij')).reshape(self.dimensions, -1).T
         hist_ravel = self.histogram.ravel()
-        hist_ravel = hist_ravel.astype(np.float) / np.nansum(hist_ravel)
+        hist_ravel = hist_ravel.astype(np.float) 
+        hist_ravel = hist_ravel / np.nansum(hist_ravel)
         result = bin_centers_ravel[np.random.choice(len(bin_centers_ravel),
                                                     p=hist_ravel,
                                                     size=size)]
@@ -650,7 +739,7 @@ class Histdd(MultiHistBase):
                                 len(self.bin_edges[i]) - 2)
                         for i in range(self.dimensions)]
         # Use the index arrays to slice the histogram
-        return self.histogram[index_arrays]
+        return self.histogram[tuple(index_arrays)]
 
         # Check against slow version:
         # def hist_to_interpolator_slow(mh):
@@ -660,8 +749,23 @@ class Histdd(MultiHistBase):
         # y = np.random.uniform(0, 200, 100)
         # hist_to_interpolator(mh)(x, y) - hist_to_interpolator_slow(mh)(x, y)
 
+    def lookup_hist(self, mh):
+        """Return histogram within binning of Histdd mh, with values looked up in this histogram.
+
+        This is not rebinning: no interpolation /renormalization is performed.
+        It's just a lookup.
+        """
+        result = mh.similar_blank_histogram()
+        points = np.stack([mh.all_axis_bin_centers(i)
+                           for i in range(mh.dimensions)]).reshape(mh.dimensions, -1)
+        values = self.lookup(*points)
+        result.histogram = values.reshape(result.histogram.shape)
+        return result
+
     def plot(self, log_scale=False, log_scale_vmin=1,
-             cblabel='Number of entries', colorbar_kwargs=None,
+             colorbar=True,
+             cblabel='Number of entries',
+             colorbar_kwargs=None,
              plt=plt,
              **kwargs):
 
@@ -682,12 +786,14 @@ class Histdd(MultiHistBase):
             mesh = plt.pcolormesh(self.bin_edges[0], self.bin_edges[1], self.histogram.T, **kwargs)
             plt.xlim(np.min(self.bin_edges[0]), np.max(self.bin_edges[0]))
             plt.ylim(np.min(self.bin_edges[1]), np.max(self.bin_edges[1]))
-            cb = plt.colorbar(**colorbar_kwargs)
-            cb.ax.minorticks_on()
             if self.axis_names:
                 plt.xlabel(self.axis_names[0])
                 plt.ylabel(self.axis_names[1])
-            return mesh, cb
+            if colorbar:
+                cb = plt.colorbar(**colorbar_kwargs)
+                cb.ax.minorticks_on()
+                return mesh, cb
+            return mesh
 
         else:
             raise ValueError("Can only plot 1- or 2-dimensional histograms!")
@@ -738,3 +844,64 @@ if __name__ == '__main__':
 
         plt.legend()
         plt.show()
+
+
+##
+# Error bar helpers
+##
+
+# Zero-background 1 sigma Poisson Feldman-Cousins intervals
+# From table II in https://arxiv.org/pdf/physics/9711021.pdf
+_fc_intervals = np.array([
+    [0.0, 1.29],
+    [0.37, 2.75],
+    [0.74, 4.25],
+    [1.1, 5.3],
+    [2.34, 6.78],
+    [2.75, 7.81],
+    [3.82, 9.28],
+    [4.25, 10.3],
+    [5.3, 11.32],
+    [6.44, 12.79],
+    [6.78, 13.81],
+    [7.81, 14.82],
+    [8.83, 16.29],
+    [9.28, 17.3],
+    [10.3, 18.32],
+    [11.32, 19.32],
+    [12.33, 20.8],
+    [12.79, 21.81],
+    [13.81, 22.82],
+    [14.82, 23.82],
+    [15.83, 25.3]])
+
+
+def poisson_central_interval(k, cl=0.6826894921370859):
+    """Return central Poisson confidence interval
+    :param k: observed events
+    :param cl: confidence level
+    """
+    if not HAVE_SCIPY:
+        raise NotImplementedError("Poisson errors require scipy")
+    # Adapted from https://stackoverflow.com/a/14832525
+    k = np.asarray(k).astype(np.int)
+    alpha = 1 - cl
+    low = stats.chi2.ppf(alpha / 2, 2 * k) / 2
+    high = stats.chi2.ppf(1 - alpha / 2, 2 * k + 2) / 2
+    return np.stack([np.nan_to_num(low), high])
+
+
+def poisson_1s_interval(k, fc=True):
+    """Return (low, high) 1 sigma Poisson confidence intervals
+
+    :param k: Observed events (int or array of ints)
+    :param fc: if True (default), use Feldman-Cousins for k <= 20,
+    and central intervals otherwise.
+    (at k = 20, the difference between these is 1-2%).
+    """
+    k = np.asarray(k).astype(np.int)
+    result = poisson_central_interval(k)
+    if fc:
+        mask = k <= 20
+        result[:, mask] = _fc_intervals[k[mask]].T
+    return result
